from aegis_gui.pages.tab_plot.plot.prep_setup import FIG_SETUP
import plotly.graph_objs as go

bootstrap_colors = [
    "#007bff",  # primary
    "#6c757d",  # secondary
    "#28a745",  # success
    "#dc3545",  # danger
    "#ffc107",  # warning
    "#17a2b8",  # info
    "#f8f9fa",  # light
    "#343a40",  # dark
]

FIG_xaxis = dict(
    # showgrid=False,
    zeroline=False,
    # showline=False,
    nticks=10,
    # gridcolor="#cccccc",
    griddash="dot",
    # gridwidth=1,
    linecolor="rgb(46, 49, 51)",
)

FIG_yaxis = dict(
    # showgrid=False,
    zeroline=False,
    # showline=False,
    nticks=10,
    # gridcolor="#cccccc",
    griddash="dot",
    # gridwidth=1,
    linecolor="rgb(46, 49, 51)",
)

FIG_legend = dict(x=0.97, y=1, yanchor="top", xanchor="right", orientation="v")

FIG_LAYOUT = dict(
    width=300 * 1.1,
    height=300 * 1.1,
    # margin={"t": 0, "r": 0, "b": 0, "l": 0},
    margin={"t": 0, "r": 0, "b": 0, "l": 0},
    plot_bgcolor="rgba(150, 150, 200, 0.1)",
    paper_bgcolor="rgba(24, 25, 27, 0.0)",
    # font_color="white",
    # showlegend=True,
    # legend=dict(x=1.05, y=1),
    # showlegend=False,
    showlegend=True,
    # legend=dict(x=0, y=0),
    legend=FIG_legend,
    font_size=14,
    # font_family="Inter",
    xaxis=FIG_xaxis,
    yaxis=FIG_yaxis,
)

dark_mode_white = "rgba(255,255,255,0.8)"


FIG_LAYOUT_DARK_MODE = dict(
    plot_bgcolor="rgba(0, 0, 0, 0.1)",  # Darker background for the plot area
    paper_bgcolor="rgba(24, 25, 27, 0.0)",  # Darker background for the overall figure
    font_color=dark_mode_white,  # Light font color for better contrast
    legend={**FIG_legend, **dict(font=dict(color=dark_mode_white))},  # Light legend text
    xaxis={
        **FIG_xaxis,
        **dict(
            linecolor=dark_mode_white,  # Light color for x-axis line
            tickfont=dict(color=dark_mode_white),  # Light color for x-axis ticks
            gridcolor="rgba(255, 255, 255, 0.1)",  # Light grid lines for visibility
        ),
    },
    yaxis={
        **FIG_yaxis,
        **dict(
            linecolor=dark_mode_white,  # Light color for y-axis line
            tickfont=dict(color=dark_mode_white),  # Light color for y-axis ticks
            gridcolor="rgba(255, 255, 255, 0.1)",  # Light grid lines for visibility
        ),
    },
)


def make_empty_figure(dark_mode=False):
    return go.Figure(layout=go.Layout({**FIG_LAYOUT, **(FIG_LAYOUT_DARK_MODE if dark_mode else {})}))


def make_scatter_figure(id_, xs, ys, selected_sims, dark_mode=False):
    figure = go.Figure(
        data=[
            go.Scatter(
                x=x, y=y, mode="markers", name=sim, marker=dict(color=bootstrap_colors[i % len(bootstrap_colors)])
            )
            for i, (x, y, sim) in enumerate(zip(xs, ys, selected_sims))
        ],
        layout=go.Layout(
            {**FIG_LAYOUT, **(FIG_LAYOUT_DARK_MODE if dark_mode else {}), **FIG_SETUP[id_]["figure_layout"]}
        ),
    )

    # Compute maximum axes limits
    maxx = max(max(x) for x in xs) if ys else 1
    maxy = max(max(y) for y in ys) if ys else 1

    # Update axes
    figure.update_xaxes(
        range=[0, 1 if maxx < 1 else maxx * 1.05],
    )
    figure.update_yaxes(
        range=[0, 1.05 if maxy < 1 else maxy * 1.1],
    )

    # Custom plots
    if id_ == "birth table":
        figure.update_yaxes(range=[0, maxy * 1.05])

    return figure


def make_line_figure(id_, xs, ys, selected_sims, dark_mode=False):
    figure = go.Figure(
        data=[
            go.Scatter(x=x, y=y, mode="lines", name=sim, line=dict(color=bootstrap_colors[i % len(bootstrap_colors)]))
            for i, (x, y, sim) in enumerate(zip(xs, ys, selected_sims))
        ],
        layout=go.Layout(
            {**FIG_LAYOUT, **(FIG_LAYOUT_DARK_MODE if dark_mode else {}), **FIG_SETUP[id_]["figure_layout"]}
        ),
    )

    # Compute maximum axes limits
    maxx = max(max(x) for x in xs) if xs else 1
    maxy = max(max(y) for y in ys) if ys else 1

    # Update axes
    figure.update_xaxes(
        range=[0, 1 if maxx < 1 else maxx * 1.05],
    )
    figure.update_yaxes(
        range=[0, 1.05 if maxy < 1 else maxy * 1.1],
    )

    # Custom plots
    if id_ == "birth table":
        figure.update_yaxes(range=[0, maxy * 1.05])

    return figure


def make_hist_figure(id_, xs, ys, selected_sims, dark_mode=False):
    fig_setup = FIG_SETUP[id_]
    figure = go.Figure(
        data=[
            go.Histogram(
                x=y,
                name=sim,
                nbinsx=fig_setup["nbinsx"] if "nbinsx" in fig_setup else None,
                marker_color=bootstrap_colors[i % len(bootstrap_colors)],
            )
            for i, (y, sim) in enumerate(zip(ys, selected_sims))
        ],
        layout=go.Layout(
            {**FIG_LAYOUT, **(FIG_LAYOUT_DARK_MODE if dark_mode else {}), **FIG_SETUP[id_]["figure_layout"]}
        ),
    )

    return figure


def make_heatmap_figure(id_, xs, ys, selected_sims, dark_mode=False):
    x = xs[0]
    y = ys[0]
    sim = next(iter(selected_sims))
    figure = go.Figure(
        data=go.Heatmap(z=y.T, x=x, name=sim, colorscale="Electric", showscale=True),
        layout=go.Layout(
            {**FIG_LAYOUT, **(FIG_LAYOUT_DARK_MODE if dark_mode else {}), **FIG_SETUP[id_]["figure_layout"]}
        ),
    )

    figure.update_yaxes(
        autorange="reversed",
        # ticks="",
        # showticklabels=False,
    )

    return figure


def make_bar_figure_stacked(id_, xs, ys, selected_sims, dark_mode=False):
    assert len(ys) > 0, f"{id_}"
    y = ys[0]
    figure = go.Figure(
        data=[
            go.Bar(x=y.index, y=y.loc[:, i], width=0.7, name=i, marker=dict(line=dict(width=0 if dark_mode else 0)))
            for i in y.columns
        ],
        layout=go.Layout(
            {**FIG_LAYOUT, **(FIG_LAYOUT_DARK_MODE if dark_mode else {}), **FIG_SETUP[id_]["figure_layout"]},
            barmode="stack",
            showlegend=True,
            # width=420,
        ),
    )
    return figure


def make_bar_figure_not_stacked(id_, xs, ys, selected_sims, dark_mode=False):

    figure = go.Figure(
        data=[
            go.Bar(x=x, y=y, name=sim, marker=dict(line=dict(width=0 if dark_mode else 0)))
            # for i in columns
            for i, (x, y, sim) in enumerate(zip(xs, ys, selected_sims))
        ],
        layout=go.Layout(
            {
                **FIG_LAYOUT,
                **(FIG_LAYOUT_DARK_MODE if dark_mode else {}),
                **FIG_SETUP[id_]["figure_layout"],
            },
            barmode="group",
            showlegend=True,
            bargroupgap=0.1,
            bargap=0.15,
            # width=420,
        ),
    )
    return figure



import math
import os


def get_folder_size_with_du(folder_path):
    """Returns the folder size for Linux, macOS, and Windows."""

    size = get_folder_size_python(folder_path)
    return size


def convert_size(size_bytes):
    """Converts bytes to a human-readable size format (e.g., KB, MB, GB)."""
    if size_bytes == 0:
        return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return f"{s} {size_name[i]}"


def get_folder_size_python(folder_path):
    """Returns the total size of the folder in a human-readable format."""
    total_size = 0

    # Traverse the directory tree
    for dirpath, dirnames, filenames in os.walk(folder_path):
        for filename in filenames:
            # Get the full file path
            filepath = os.path.join(dirpath, filename)
            # Skip if it's a symbolic link
            if not os.path.islink(filepath):
                # Accumulate file size
                total_size += os.path.getsize(filepath)

    return convert_size(total_size)



import pathlib
import yaml
from .recorder import Recorder


class ConfigRecorder(Recorder):
    """

    Records once.
    """

    def __init__(self, odir: pathlib.Path):
        self.odir = odir
        self.path = self.odir / "final_config.yml"

    def write_final_config_file(self, final_config):
        with open(self.path, "w") as file_:
            yaml.dump(final_config, file_)



import logging
import dash_bootstrap_components as dbc
import dash
import pandas as pd
from dash import callback, Output, Input, ctx, State, MATCH, dcc, html
from plotly.io import write_image
from aegis_gui.guisettings.GuiSettings import gui_settings

# FIGURE DOWNLOAD


@callback(
    Output({"type": "figure-dcc-download", "index": MATCH}, "data"),
    Input({"type": "figure-download-button", "index": MATCH}, "n_clicks"),
    State({"type": "graph-figure", "index": MATCH}, "figure"),
)
def figure_download_button_click(n_clicks, figure):

    if n_clicks is None:
        return

    fig_name = ctx.triggered_id["index"]
    path_figure = gui_settings.figure_dir / f"{fig_name}.png"
    logging.info(f"Recording figure to {path_figure}...")
    write_image(figure, path_figure)
    logging.info(f"'{fig_name}' has been recorded successfully.")

    return dcc.send_file(path_figure)


def get_figure_download_button(figure_id):
    button = dbc.Button(
        [html.I(className="bi bi-file-arrow-down-fill"), "Save figure"],
        id={"type": "figure-download-button", "index": figure_id},
    )
    return button


def get_figure_download_dcc(figure_id):
    return dcc.Download(id={"type": "figure-dcc-download", "index": figure_id})


# DATA DOWNLOAD


@callback(
    Output({"type": "data-dcc-download", "index": MATCH}, "data"),
    Input({"type": "data-download-button", "index": MATCH}, "n_clicks"),
)
def download_data(n_clicks):
    if n_clicks is None:
        return
    # Convert the sample data to a DataFrame
    df = pd.DataFrame([1, 2, 3, 4])

    # Convert DataFrame to CSV
    csv_string = df.to_csv(index=False)

    return dcc.send_data_frame(df.to_csv, "data.csv")


def get_data_download_button(figure_id):
    button = dbc.Button(
        [html.I(className="bi bi-file-arrow-down-fill"), "Download figure data"],
        id={"type": "data-download-button", "index": figure_id},
        disabled=True,
    )
    return button


def get_data_download_dcc(figure_id):
    return dcc.Download(id={"type": "data-dcc-download", "index": figure_id})



import dash
from dash import html, dcc
import dash_bootstrap_components as dbc
from aegis_gui.utilities import sim_tracker
from aegis_gui.utilities.utilities import get_icon
from aegis_gui import offcanvas
from aegis_gui.guisettings.GuiSettings import gui_settings


def get_app_layout():
    return html.Div(
        id="body-container",
        children=[
            dcc.Location(id="url", refresh=False),
            get_sidebar(),
            html.Div(id="main-container", children=[dash.page_container]),
        ],
    )


def get_sidebar():
    return dash.html.Div(
        [
            dbc.Nav(
                children=[
                    html.A(
                        [
                            html.Img(
                                src="assets/aegis-ager.svg", width="80%", style={"margin": "0 10%"}, id="aegis-logo"
                            )
                        ],
                        href=gui_settings.wrap_href(""),
                        className="mb-5",
                    ),
                    dbc.NavItem(
                        [
                            dbc.NavLink(
                                [
                                    dash.html.I(className="bi bi-house-door-fill"),
                                    "Home",
                                    # get_icon("house-door-fill"),
                                ],
                                href=gui_settings.wrap_href(""),
                                id="link-nav-home",
                            )
                        ]
                    ),
                    dbc.NavItem(
                        [
                            dbc.NavLink(
                                [dash.html.I(className="bi bi-rocket-takeoff-fill"), "Launch"],
                                href=gui_settings.wrap_href("config"),
                                id="link-nav-config",
                            )
                        ]
                    ),
                    dbc.NavItem(
                        [
                            dbc.NavLink(
                                [dash.html.I(className="bi bi-bar-chart-fill"), "Plot"],
                                href=gui_settings.wrap_href("plot"),
                                id="link-nav-plot",
                            )
                        ]
                    ),
                    dbc.NavItem(
                        [
                            dbc.NavLink(
                                [dash.html.I(className="bi bi-eye-fill"), "Control"],
                                href=gui_settings.wrap_href("simlog"),
                                id="link-nav-simlog",
                            )
                        ]
                    ),
                    dbc.NavItem(
                        [
                            dbc.NavLink(
                                [dash.html.I(className="bi bi-info-square-fill"), "Wiki"],
                                href=gui_settings.wrap_href("wiki"),
                                id="link-nav-wiki",
                            )
                        ]
                    ),
                ]
                + sim_tracker.init_tracker_box()
                + [offcanvas.get_offcanvas_trigger(), offcanvas.get_offcanvas()],
                id="sidebar",
                vertical="md",
                pills=True,  # TODO fix because it is not showing since i changed the href logic with gui_settings.wrap_href
                # fill=True,
            ),
        ],
    )



from aegis_gui.pages.tab_plot.plot.prep_setup import FIG_SETUP
from aegis_gui.pages.tab_plot.plot import prep_fig
from aegis_gui.utilities.utilities import get_container


def generate_figure(fig_name, selected_sims, iloc, dark_mode=False):
    """Generates a figure using the figure setup"""

    # Extract setup
    fig_setup = FIG_SETUP[fig_name]

    # Prepare x and y data
    prep_x = fig_setup["prep_x"]
    prep_y = fig_setup["prep_y"]

    ys = []
    max_ilocs = []
    xs = []

    # TODO you might not have to read all the data again when you slide!

    for sim in selected_sims:
        container = get_container(sim)
        ysi, max_iloc = prep_y(container, iloc=iloc)
        ys.append(ysi)

        if max_iloc is not None:
            max_iloc -= 1  # before this, it is length
            max_ilocs.append(max_iloc)

        xsi = prep_x(container, y=ysi)
        xs.append(xsi)

    max_iloc = min(max_ilocs) if max_ilocs else None

    prep_figure = getattr(prep_fig, fig_setup["prep_figure"])
    figure = prep_figure(fig_name, xs, ys, selected_sims, dark_mode=dark_mode)

    return figure, max_iloc



import dash
import dash_bootstrap_components as dbc
from aegis_gui.utilities import log

seconds_per_refresh = 10
progress_granularity = 0.01  # 1% of the progress bar
seconds_per_progress_step = seconds_per_refresh * progress_granularity


inputgroup = dash.html.Div(
    [
        dbc.Button(
            [dash.html.I(className="bi bi-arrow-clockwise"), "Refresh"],
            id="refresh-figure-data",
            className="me-1",
            outline=True,
            color="secondary",
        ),
        dbc.Button(
            [dash.html.I(className="bi bi-arrow-repeat"), "Keep refreshing"],
            id="keep-refresh-figure-data",
            outline=True,
            color="secondary",
            className="me-1",
            active=False,
        ),
        dash.html.Div(
            children=[
                dbc.Progress(
                    id="refresh-progress-bar",
                    value=0,
                    # striped=True,
                    # animated=True,
                    style={
                        # "height": "8px",
                        "width": "38px",
                        "transform": "rotate(-90deg)",
                        "transformOrigin": "top left",
                        "translate": "0 38px",
                    },
                )
            ],
            className="me-1",
            style={
                # "height": "8px",
                # "width": "38px",
            },
        ),
        dash.dcc.Interval(
            id="progress-interval-refresh",
            interval=1000 * seconds_per_progress_step,
            n_intervals=0,
            disabled=True,
        ),
        dash.dcc.Store(
            id="refresh-progress-store",
            data={"progress": 0},
        ),
    ],
    style={"display": "flex"},
)

# TODO probably unify these two because the progressbar is not going to 0 sometimes when button is unactived


@dash.callback(
    dash.Output("progress-interval-refresh", "disabled"),
    dash.Output("keep-refresh-figure-data", "active"),
    dash.Output("refresh-progress-store", "data", allow_duplicate=True),
    dash.Input("keep-refresh-figure-data", "n_clicks"),
    dash.State("progress-interval-refresh", "disabled"),
    dash.State("refresh-progress-store", "data"),
    prevent_initial_call=True,
)
@log.log_debug
def toggle_interval(n_clicks, is_disabled, progress_data):
    if n_clicks is None:
        raise dash.no_update

    if is_disabled:
        # If the interval is disabled, enable it and reset progress
        return False, True, {"progress": 0}
    else:
        # Otherwise, disable the interval and stop progress
        return True, False, {"progress": 0}


@dash.callback(
    dash.Output("refresh-progress-bar", "value"),
    dash.Output("refresh-progress-store", "data", allow_duplicate=True),
    dash.Output("refresh-figure-data", "n_clicks"),
    dash.Input("progress-interval-refresh", "n_intervals"),
    dash.State("refresh-progress-store", "data"),
    dash.State("refresh-figure-data", "n_clicks"),
    prevent_initial_call=True,
)
def update_progress(n_intervals, progress_data, n_clicks):
    progress = progress_data["progress"]

    # Increment the progress by 10% each second (adjust as needed)
    progress += progress_granularity * 100

    if progress > 100:
        return 0, {"progress": 0}, 1 if n_clicks is None else n_clicks + 1
    else:
        return progress, {"progress": progress}, dash.no_update



import dash
from aegis_sim.utilities.container import Container
from aegis_gui.utilities import utilities
import dash_bootstrap_components as dbc
from aegis_gui.utilities import log
from aegis_gui.guisettings.GuiSettings import gui_settings


def make_trackers(ticker_store):
    trackers = []

    running_sims = []
    paths = utilities.get_sim_paths()
    for path in paths:

        ticker_store_status = ticker_store.get(path.stem)
        ticker_registered_and_finished = ticker_store_status is False
        if ticker_registered_and_finished:
            continue

        container = Container(path)
        is_running = not container.has_ticker_stopped()
        ticker_store[path.stem] = is_running
        if is_running:
            progress = container.get_simple_log()
            if progress is None:
                continue
            progress = progress[0] / progress[1]
            progressbar = dbc.Progress(
                label=container.name,
                value=10 + progress * 90,
                striped=True,
                animated=True,
                style={"marginTop": "0.5rem"},
            )
            linked_progressbar = dash.dcc.Link(
                children=progressbar,
                href=gui_settings.wrap_href(f"simlog?sim={container.name}"),
            )
            # running_sims.append(tracker)
            running_sims.append(linked_progressbar)

    if running_sims:
        trackers.append(
            dbc.Toast(
                children=[
                    # dash.html.Hr(),
                    # dash.html.P("Running simulations", className="text-secondary", style={"margin": 0}),
                ]
                + running_sims,
                header="Running sims",
                style={"width": "100%", "marginTop": "1rem"},
            )
        )

    recent_containers = []

    RECENCY_WINDOW = 365 * 24 * 3600

    for path in paths:
        container = Container(path)
        ticker = container.get_ticker()
        since_last = ticker.since_last()
        if since_last is None:
            continue
        if since_last < RECENCY_WINDOW and container.has_ticker_stopped():
            recent_containers.append(container)

    if recent_containers:
        N_recent_containers_to_show = 20

        buttons = [
            dbc.Button(
                children=[rc.name],
                href=gui_settings.wrap_href(f"simlog?sim={rc.name}"),
                className="badge me-1",
                color="primary",
            )
            for rc in recent_containers[:N_recent_containers_to_show]
        ]

        if len(recent_containers) > N_recent_containers_to_show:
            buttons.append(
                dbc.Button("more...", className="badge me-1", color="secondary", href=gui_settings.wrap_href(f"simlog"))
            )

        nav = dash.html.Div(children=buttons)

        trackers.append(
            dbc.Toast(
                children=nav,
                header="Most recent sims",
                style={"width": "100%", "marginTop": "1rem"},
            )
        )

        # trackers.append(nav)

    return trackers, ticker_store


@log.log_debug
def init_tracker_box():
    trackers, ticker_store = get_tracker_box(None, {})
    return [
        dash.dcc.Interval(
            id="running-simulations-interval", interval=2 * 1000, n_intervals=0
        ),  # TODO potential performance issues
        dash.dcc.Store(id="ticker-store", data=ticker_store),
        dash.html.Div(trackers, id="running-simulations"),
    ]


@dash.callback(
    dash.Output("running-simulations", "children"),
    dash.Output("ticker-store", "data"),
    dash.Input("running-simulations-interval", "n_intervals"),
    dash.State("ticker-store", "data"),
)
def get_tracker_box(n, ticker_store):
    # NOTE This takes a long time. If the interval refresh rate is very high, it will be retriggered before finishing and no trackers will appear.
    trackers, ticker_store = make_trackers(ticker_store)
    return trackers, ticker_store
